@@ -7,7 +7,7 @@ import threading
 from datetime import datetime, timedelta
 from gzip import GzipFile
 from tempfile import SpooledTemporaryFile
-from urllib.parse import parse_qsl, urlsplit
+from urllib.parse import parse_qsl, urlencode, urlsplit
 
 from django.contrib.staticfiles.storage import ManifestFilesMixin
 from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
@@ -42,6 +42,8 @@ def _use_cryptography_signer():
     )
 
     def _cloud_front_signer_from_pem(key_id, pem):
+        if isinstance(pem, str):
+            pem = pem.encode('ascii')
         key = load_pem_private_key(
             pem, password=None, backend=default_backend())
 
@@ -56,6 +58,8 @@ def _use_rsa_signer():
     import rsa
 
     def _cloud_front_signer_from_pem(key_id, pem):
+        if isinstance(pem, str):
+            pem = pem.encode('ascii')
         key = rsa.PrivateKey.load_pkcs1(pem)
         return CloudFrontSigner(key_id, lambda x: rsa.sign(x, key, 'SHA-1'))
 
@@ -109,6 +113,7 @@ class S3Boto3StorageFile(File):
         self._raw_bytes_written = 0
         self._file = None
         self._multipart = None
+        self._parts = None
         # 5 MB is the minimum part size (if there is more than one part).
         # Amazon allows up to 10,000 parts.  The default supports uploads
         # up to roughly 50 GB.  Increase the part size to accommodate
@@ -158,6 +163,7 @@ class S3Boto3StorageFile(File):
             self._multipart = self.obj.initiate_multipart_upload(
                 **self._storage._get_write_parameters(self.obj.key)
             )
+            self._parts = []
         if self.buffer_size <= self._buffer_file_size:
             self._flush_write_buffer()
         bstr = to_bytes(content)
@@ -180,7 +186,11 @@ class S3Boto3StorageFile(File):
             self._write_counter += 1
             self.file.seek(0)
             part = self._multipart.Part(self._write_counter)
-            part.upload(Body=self.file.read())
+            response = part.upload(Body=self.file.read())
+            self._parts.append({
+                'ETag': response['ETag'],
+                'PartNumber': self._write_counter
+            })
             self.file.seek(0)
             self.file.truncate()
 
@@ -212,13 +222,8 @@ class S3Boto3StorageFile(File):
     def close(self):
         if self._is_dirty:
             self._flush_write_buffer()
-            # TODO: Possibly cache the part ids as they're being uploaded
-            # instead of requesting parts from server. For now, emulating
-            # s3boto's behavior.
-            parts = [{'ETag': part.e_tag, 'PartNumber': part.part_number}
-                     for part in self._multipart.parts.all()]
             self._multipart.complete(
-                MultipartUpload={'Parts': parts})
+                MultipartUpload={'Parts': self._parts})
         else:
             if self._multipart is not None:
                 self._multipart.abort()
@@ -289,9 +294,19 @@ class S3Boto3Storage(BaseStorage):
         else:
             cloudfront_signer = None
 
+        s3_access_key_id = setting('AWS_S3_ACCESS_KEY_ID')
+        s3_secret_access_key = setting('AWS_S3_SECRET_ACCESS_KEY')
+        s3_session_profile = setting('AWS_S3_SESSION_PROFILE')
+        if (s3_access_key_id or s3_secret_access_key) and s3_session_profile:
+            raise ImproperlyConfigured(
+                'AWS_S3_SESSION_PROFILE should not be provided with '
+                'AWS_S3_ACCESS_KEY_ID and AWS_S3_SECRET_ACCESS_KEY'
+            )
+
         return {
             "access_key": setting('AWS_S3_ACCESS_KEY_ID', setting('AWS_ACCESS_KEY_ID')),
             "secret_key": setting('AWS_S3_SECRET_ACCESS_KEY', setting('AWS_SECRET_ACCESS_KEY')),
+            "session_profile": setting('AWS_S3_SESSION_PROFILE'),
             "file_overwrite": setting('AWS_S3_FILE_OVERWRITE', True),
             "object_parameters": setting('AWS_S3_OBJECT_PARAMETERS', {}),
             "bucket_name": setting('AWS_STORAGE_BUCKET_NAME'),
@@ -337,12 +352,9 @@ class S3Boto3Storage(BaseStorage):
     def connection(self):
         connection = getattr(self._connections, 'connection', None)
         if connection is None:
-            session = boto3.session.Session()
+            session = self._create_session()
             self._connections.connection = session.resource(
                 's3',
-                aws_access_key_id=self.access_key,
-                aws_secret_access_key=self.secret_key,
-                aws_session_token=self.security_token,
                 region_name=self.region_name,
                 use_ssl=self.use_ssl,
                 endpoint_url=self.endpoint_url,
@@ -351,6 +363,22 @@ class S3Boto3Storage(BaseStorage):
             )
         return self._connections.connection
 
+    def _create_session(self):
+        """
+        If a user specifies a profile name and this class obtains access keys
+        from another source such as environment variables,we want the profile
+        name to take precedence.
+        """
+        if self.session_profile:
+            session = boto3.Session(profile_name=self.session_profile)
+        else:
+            session = boto3.Session(
+                    aws_access_key_id=self.access_key,
+                    aws_secret_access_key=self.secret_key,
+                    aws_session_token=self.security_token
+            )
+        return session
+
     @property
     def bucket(self):
         """
@@ -367,8 +395,8 @@ class S3Boto3Storage(BaseStorage):
         provided in the settings then get them from the environment
         variables.
         """
-        access_key = self.access_key or lookup_env(S3Boto3Storage.access_key_names)
-        secret_key = self.secret_key or lookup_env(S3Boto3Storage.secret_key_names)
+        access_key = self.access_key or lookup_env(self.access_key_names)
+        secret_key = self.secret_key or lookup_env(self.secret_key_names)
         return access_key, secret_key
 
     def _get_security_token(self):
@@ -376,7 +404,7 @@ class S3Boto3Storage(BaseStorage):
         Gets the security token to use when accessing S3. Get it from
         the environment variables.
         """
-        security_token = self.security_token or lookup_env(S3Boto3Storage.security_token_names)
+        security_token = self.security_token or lookup_env(self.security_token_names)
         return security_token
 
     def _clean_name(self, name):
@@ -407,7 +435,6 @@ class S3Boto3Storage(BaseStorage):
 
     def _compress_content(self, content):
         """Gzip a given string content."""
-        content.seek(0)
         zbuf = io.BytesIO()
         #  The GZIP header has a modification time attribute (see http://www.zlib.org/rfc-gzip.html)
         #  This means each time a file is compressed it changes even if the other contents don't change
@@ -436,6 +463,8 @@ class S3Boto3Storage(BaseStorage):
         name = self._normalize_name(cleaned_name)
         params = self._get_write_parameters(name, content)
 
+        if content.seekable():
+            content.seek(0, os.SEEK_SET)
         if (self.gzip and
                 params['ContentType'] in self.gzip_content_types and
                 'ContentEncoding' not in params):
@@ -443,7 +472,6 @@ class S3Boto3Storage(BaseStorage):
             params['ContentEncoding'] = 'gzip'
 
         obj = self.bucket.Object(name)
-        content.seek(0, os.SEEK_SET)
         obj.upload_fileobj(content, ExtraArgs=params)
         return cleaned_name
 
@@ -456,8 +484,14 @@ class S3Boto3Storage(BaseStorage):
         try:
             self.connection.meta.client.head_object(Bucket=self.bucket_name, Key=name)
             return True
-        except ClientError:
-            return False
+        except ClientError as error:
+            if error.response.get('Error', {}).get('Code') == '404':
+                return False
+
+            # Some other error was encountered. As `get_available_name` calls this,
+            # we have to assume the filename is unavailable. If we return true due to some
+            # other error, we'd overwrite a file.
+            return True
 
     def listdir(self, name):
         path = self._normalize_name(self._clean_name(name))
@@ -474,7 +508,9 @@ class S3Boto3Storage(BaseStorage):
             for entry in page.get('CommonPrefixes', ()):
                 directories.append(posixpath.relpath(entry['Prefix'], path))
             for entry in page.get('Contents', ()):
-                files.append(posixpath.relpath(entry['Key'], path))
+                key = entry['Key']
+                if key != path:
+                    files.append(posixpath.relpath(key, path))
         return directories, files
 
     def size(self, name):
@@ -554,12 +590,17 @@ class S3Boto3Storage(BaseStorage):
     def url(self, name, parameters=None, expire=None, http_method=None):
         # Preserve the trailing slash after normalizing the path.
         name = self._normalize_name(self._clean_name(name))
+        params = parameters.copy() if parameters else {}
         if expire is None:
             expire = self.querystring_expire
 
         if self.custom_domain:
-            url = "{}//{}/{}".format(
-                self.url_protocol, self.custom_domain, filepath_to_uri(name))
+            url = "{}//{}/{}{}".format(
+                self.url_protocol,
+                self.custom_domain,
+                filepath_to_uri(name),
+                "?{}".format(urlencode(params)) if params else "",
+            )
 
             if self.querystring_auth and self.cloudfront_signer:
                 expiration = datetime.utcnow() + timedelta(seconds=expire)
@@ -568,7 +609,6 @@ class S3Boto3Storage(BaseStorage):
 
             return url
 
-        params = parameters.copy() if parameters else {}
         params['Bucket'] = self.bucket.name
         params['Key'] = name
         url = self.bucket.meta.client.generate_presigned_url('get_object', Params=params,
